/* ==========================================================================
 * login_bsd - Shell wrapper for BSD Authentication using high-level C API
 * --------------------------------------------------------------------------
 * Copyright (c) 2003  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */

#include <stdio.h>
#include <stdlib.h>

#include <login_cap.h>	/* auth_userokay, ... */
#include <bsd_auth.h>	/* auth_userokay, ... */

#include <unistd.h>	/* getopt, sleep */

#include <sys/types.h>	/* dbopen, md5 */

#include <fcntl.h>	/* dbopen */
#include <limits.h>	/* dbopen */
#include <db.h>		/* dbopen */
#include <md5.h>

#include <syslog.h>	/* syslog */

#include <time.h>	/* time_t */

#include <string.h>	/* strcmp, strchr, strcspn, memcpy */

#include <libgen.h>	/* basename */

#include <err.h>	/* errn, warn, ... */

#include <errno.h>


#define XSTR(x) #x
#define STR(x)  XSTR(x)


#define USAGE	\
	"bsdauth [-t type] [-s service] [-v name=value] user[:style] [class]\n"	\
	"  -t type        auth type, i.e. auth-ftp\n"				\
	"  -s service     auth protocol: login | challenge\n"			\
	"  -v name=value  name/value pairs to pass to auth script\n"		\
	"  -h             display this help and exit\n"				\
	"\n"									\
	"Report bugs to William Ahern <william@25thandClement.com>"

/*
 * Program name (probably login_bsd)
 */
char *program;


/*
 * Failure Logging
 */

#ifndef FAIL_LOG
#error "FAIL_LOG must be a path to the failure log database: /var/run/login_bsd.db"
#endif

#ifndef FAIL_GRACE
#define FAIL_GRACE	3
#endif

#ifndef FAIL_EXPIRE
#define FAIL_EXPIRE	60	/* seconds */
#endif

#ifndef FAIL_PENALTY
#define FAIL_PENALTY	5
#endif


struct failure {
	time_t last;
	unsigned failures;
};


void fail_user(char *user) {
	extern char *program;
	unsigned userln;
	MD5_CTX ctx;
	unsigned char digest[16];
	DB *dp		= NULL;
	DBT key;
	DBT data;
	int ret;
	time_t cur;
	unsigned failures	= FAIL_GRACE + 1;
	struct failure fail;


	/* we're important enough to warrant attention */
	openlog(program,LOG_PERROR,LOG_AUTH);


	cur	= time(NULL);
	

	/* operate on digest to prevent algorithmic complexity attack:
	 * 	http://www.cs.rice.edu/~scrosby/hash/
	 */
	userln	= strcspn(user,":");	/* forget auth style info */
	MD5Init(&ctx);
	MD5Update(&ctx,(unsigned char *)user,userln);
	MD5Final(digest,&ctx);


	/* don't use O_CREAT because we don't want the user to have
	 * unfettered access to the database. the database will have
	 * to be touched upon installation
	 */
	dp	= dbopen(STR(FAIL_LOG),O_EXLOCK|O_RDWR,0,DB_HASH,NULL);

	if (!dp) {
		syslog(LOG_ALERT,"failed to open failure log[%s]: %m",STR(FAIL_LOG));
		goto finish;
	}


	key.data	= digest;
	key.size	= sizeof(digest);

	ret	= dp->get(dp,&key,&data,0);

	if (ret == -1) {
		syslog(LOG_ALERT,"failed to read failure log[%s]: %m",STR(FAIL_LOG));
		goto finish;
	} else if (ret == 0) {
		if (data.size != sizeof(fail)) {
			syslog(LOG_ALERT,"inconsistent data in failure log[%s]",STR(FAIL_LOG));
			goto finish;
		}

		memcpy(&fail,data.data,sizeof(fail));

		if (fail.last + FAIL_EXPIRE < cur) /* check to reset */
			fail.failures	= 0;

		fail.last	= cur;
		fail.failures++;
	} else {
		fail.last	= cur;
		fail.failures	= 1;	
	}


	data.data	= &fail;
	data.size	= sizeof(fail);

	ret	= dp->put(dp,&key,&data,0);

	if (ret != 0) {
		syslog(LOG_ALERT,"failed to write to failure log[%s]: %m",STR(FAIL_LOG));
		goto finish;
	}

	failures	= fail.failures;	

	if (failures > FAIL_GRACE)
		syslog(LOG_NOTICE,"authentication failures[%d] exceeded grace[%d]: %s",failures,FAIL_GRACE,user);

finish:

	if (dp) {
		if (dp->close(dp) != 0) { /* what should we do? truncate? */
			syslog(LOG_ALERT,"failed to close failure log[%s]: %m",STR(FAIL_LOG));
		}
	}

	if (failures > FAIL_GRACE)
		sleep(FAIL_PENALTY);

	return; /* void */
} /* fail_user */

/*
 * end Failure Logging
 */



/*
 * Auth type and function handler pair
 */
struct auth {
	char *service;
	int (*authenticate)(char *, char *, char *, char *);
};


int service_login(char *user, char *style, char *type, char *class) {
	/* sizeof response in login.c, obsd 3.3 */
	char passwd[1024]	= {0};
	int passln;
	int ch;
	int okay;

	(void)setvbuf(stdin,NULL,_IONBF,0);

	for (passln = 0; passln < sizeof(passwd); passln++) {
		ch	= getchar();
		if (ch == EOF || ch == '\0' || ch == '\n')
			break;
		passwd[passln]	= ch;
	}

	okay	= auth_userokay(user,style,type,passwd);

	if (okay) {
		return EXIT_SUCCESS;
	} else {
		fail_user(user);
		return EXIT_FAILURE;
	}
} /* service_login */


int service_challenge(char *user, char *style, char *type, char *class) {

	warnx("challenge service not support yet");

	return EXIT_FAILURE;
} /* service_challenge */


struct auth handlers[] = {
	{"login",service_login},
	{"challenge",service_challenge},
	{NULL,NULL}
};


int main(int argc, char *argv[]) {
	extern char *program;
	char *srvc	= "login";
	char *user	= NULL;
	char *style	= NULL;
	char *type	= NULL;
	char *class	= NULL;


	program	= basename(*argv);


/* get option arguments */ {

	extern char *optarg;
	extern int optind;
	int ch;


	opterr	= 0;
	while ((ch = getopt(argc,argv,"t:s:v:h")) != -1) {
		switch (ch) {
			case 't':
				type	= optarg;
				break;
			case 's':
				srvc	= optarg;
				break;
			case 'v':
				warnx("-v values not currently supported");
				break;
			case 'h':
			case '?':
			default:
				fprintf(stderr,"%s\n",USAGE);
				return EXIT_FAILURE;
		}
	}

	argc	-= optind;
	argv	+= optind;


	if (argc == 0)
		return (fprintf(stderr,"%s\n",USAGE),EXIT_FAILURE);

	user	= *argv;

	if ((style = strchr(user,':'))) {
		*style	= '\0';
		style++;
	}

	argc--;
	argv++;


	if (argc == 0)
		goto handler;

	class	= *argv;

	argc--;
	argv++;


/* end get option arguments */ }


/* call handler */ {

	extern struct auth handlers[];
	struct auth *handler;

handler:
	for (handler = handlers; handler->service; handler++) {
		if (strcmp(srvc,handler->service) == 0) {
			return handler->authenticate(user,style,type,class);
		}
	}
	
/* end call handler */ }

	warnx("no protocol handler for service: %s",srvc);

	return EXIT_FAILURE;
}
